/**
 * TrueReality - file_loader.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "../global/type_sizes.h"
#include "../global/config.h"
#include "../global/memory_extern.h"


#include "memory.h"
#include "registers.h"
#include "rsp_registers.h"
#include "rsp_mnemonic.h"
#include "rsp.h"



/* rsp registers */

struct rsp_reg rsp_reg;

/* rsp instructions */

static void rsp_special();
static void rsp_regimm();
static void rsp_j();
static void rsp_jal();
static void rsp_beq();
static void rsp_bne();
static void rsp_blez();
static void rsp_bgtz();
static void rsp_addi();
static void rsp_addiu();
static void rsp_slti();
static void rsp_sltiu();
static void rsp_andi();
static void rsp_ori();
static void rsp_xori();
static void rsp_lui();
static void rsp_cop0();
static void rsp_cop2();
static void rsp_lb();
static void rsp_lh();
static void rsp_lw();
static void rsp_lbu();
static void rsp_lhu();
static void rsp_sb();
static void rsp_sh();
static void rsp_sw();
static void rsp_lwc2();
static void rsp_swc2();
static void rsp_reserved();



/* rsp special instructions */

static void rsp_special_sll();
static void rsp_special_srl();
static void rsp_special_sra();
static void rsp_special_sllv();
static void rsp_special_srlv();
static void rsp_special_srav();
static void rsp_special_jr();
static void rsp_special_jalr();
static void rsp_special_break();
static void rsp_special_add();
static void rsp_special_addu();
static void rsp_special_sub();
static void rsp_special_subu();
static void rsp_special_and();
static void rsp_special_or();
static void rsp_special_xor();
static void rsp_special_nor();
static void rsp_special_slt();
static void rsp_special_sltu();
static void rsp_special_reserved();



/* rsp regimm instructions */

static void rsp_regimm_bltz();
static void rsp_regimm_bgez();
static void rsp_regimm_bltzal();
static void rsp_regimm_bgezal();
static void rsp_regimm_reserved();



/* rsp cop0 instructions */

static void rsp_cop0_mf();
static void rsp_cop0_mt();
static void rsp_cop0_reserved();



/* rsp cop2 instructions */

static void rsp_cop2_mf();
static void rsp_cop2_cf();
static void rsp_cop2_mt();
static void rsp_cop2_ct();
static void rsp_cop2_vectop();
static void rsp_cop2_reserved();



/* rsp cop2 vectop instructions */

static void rsp_cop2_vectop_vmulf();
static void rsp_cop2_vectop_vmulu();
static void rsp_cop2_vectop_vrndp();
static void rsp_cop2_vectop_vmulq();
static void rsp_cop2_vectop_vmudl();
static void rsp_cop2_vectop_vmudm();
static void rsp_cop2_vectop_vmudn();
static void rsp_cop2_vectop_vmudh();
static void rsp_cop2_vectop_vmacf();
static void rsp_cop2_vectop_vmacu();
static void rsp_cop2_vectop_vrndn();
static void rsp_cop2_vectop_vmacq();
static void rsp_cop2_vectop_vmadl();
static void rsp_cop2_vectop_vmadm();
static void rsp_cop2_vectop_vmadn();
static void rsp_cop2_vectop_vmadh();
static void rsp_cop2_vectop_vadd();
static void rsp_cop2_vectop_vsub();
static void rsp_cop2_vectop_vsut();
static void rsp_cop2_vectop_vabs();
static void rsp_cop2_vectop_vaddc();
static void rsp_cop2_vectop_vsubc();
static void rsp_cop2_vectop_vaddb();
static void rsp_cop2_vectop_vsubb();
static void rsp_cop2_vectop_vaccb();
static void rsp_cop2_vectop_vsucb();
static void rsp_cop2_vectop_vsad();
static void rsp_cop2_vectop_vsac();
static void rsp_cop2_vectop_vsum();
static void rsp_cop2_vectop_vsaw();
static void rsp_cop2_vectop_vlt();
static void rsp_cop2_vectop_veq();
static void rsp_cop2_vectop_vne();
static void rsp_cop2_vectop_vge();
static void rsp_cop2_vectop_vcl();
static void rsp_cop2_vectop_vch();
static void rsp_cop2_vectop_vcr();
static void rsp_cop2_vectop_vmrg();
static void rsp_cop2_vectop_vand();
static void rsp_cop2_vectop_vnand();
static void rsp_cop2_vectop_vor();
static void rsp_cop2_vectop_vnor();
static void rsp_cop2_vectop_vxor();
static void rsp_cop2_vectop_vnxor();
static void rsp_cop2_vectop_vrcp();
static void rsp_cop2_vectop_vrcpl();
static void rsp_cop2_vectop_vrcph();
static void rsp_cop2_vectop_vmov();
static void rsp_cop2_vectop_vrsq();
static void rsp_cop2_vectop_vrsql();
static void rsp_cop2_vectop_vrsqh();
static void rsp_cop2_vectop_vnoop();
static void rsp_cop2_vectop_vextt();
static void rsp_cop2_vectop_vextq();
static void rsp_cop2_vectop_vextn();
static void rsp_cop2_vectop_vinst();
static void rsp_cop2_vectop_vinsq();
static void rsp_cop2_vectop_vinsn();
static void rsp_cop2_vectop_reserved();



/* rsp instruction function pointer */

static rsp_instr instruction[64] =
{
        rsp_special,
        rsp_regimm,
        rsp_j,
        rsp_jal,
        rsp_beq,
        rsp_bne,
        rsp_blez,
        rsp_bgtz,
        rsp_addi,
        rsp_addiu,
        rsp_slti,
        rsp_sltiu,
        rsp_andi,
        rsp_ori,
        rsp_xori,
        rsp_lui,
        rsp_cop0,
        rsp_reserved,
        rsp_cop2,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_lb,
        rsp_lh,
        rsp_reserved,
        rsp_lw,
        rsp_lbu,
        rsp_lhu,
        rsp_reserved,
        rsp_reserved,
        rsp_sb,
        rsp_sh,
        rsp_reserved,
        rsp_sw,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_lwc2,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_swc2,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved,
        rsp_reserved
};





/* rsp special function pointer */

static rsp_instr special_instruction[64] =
{
        rsp_special_sll,
        rsp_special_reserved,
        rsp_special_srl,
        rsp_special_sra,
        rsp_special_sllv,
        rsp_special_reserved,
        rsp_special_srlv,
        rsp_special_srav,
        rsp_special_jr,
        rsp_special_jalr,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_break,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_add,
        rsp_special_addu,
        rsp_special_sub,
        rsp_special_subu,
        rsp_special_and,
        rsp_special_or,
        rsp_special_xor,
        rsp_special_nor,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_slt,
        rsp_special_sltu,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved,
        rsp_special_reserved
};



/* rsp regimm function pointer */

static rsp_instr regimm_instruction[32] =
{
        rsp_regimm_bltz,
        rsp_regimm_bgez,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_bltzal,
        rsp_regimm_bgezal,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved,
        rsp_regimm_reserved
};



/* rsp cop0 function pointer */

static rsp_instr cop0_instruction[32] =
{
        rsp_cop0_mf,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_mt,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved,
        rsp_cop0_reserved
};



/* rsp cop2 function pointer */

static rsp_instr cop2_instruction[32] =
{
        rsp_cop2_mf,
        rsp_cop2_reserved,
        rsp_cop2_cf,
        rsp_cop2_reserved,
        rsp_cop2_mt,
        rsp_cop2_reserved,
        rsp_cop2_ct,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_reserved,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop,
        rsp_cop2_vectop
};



/* rsp cop2 vectop function pointer */

static rsp_instr cop2_vectop_instruction[64] =
{
        rsp_cop2_vectop_vmulf,
        rsp_cop2_vectop_vmulu,
        rsp_cop2_vectop_vrndp,
        rsp_cop2_vectop_vmulq,
        rsp_cop2_vectop_vmudl,
        rsp_cop2_vectop_vmudm,
        rsp_cop2_vectop_vmudn,
        rsp_cop2_vectop_vmudh,
        rsp_cop2_vectop_vmacf,
        rsp_cop2_vectop_vmacu,
        rsp_cop2_vectop_vrndn,
        rsp_cop2_vectop_vmacq,
        rsp_cop2_vectop_vmadl,
        rsp_cop2_vectop_vmadm,
        rsp_cop2_vectop_vmadn,
        rsp_cop2_vectop_vmadh,
        rsp_cop2_vectop_vadd,
        rsp_cop2_vectop_vsub,
        rsp_cop2_vectop_vsut,
        rsp_cop2_vectop_vabs,
        rsp_cop2_vectop_vaddc,
        rsp_cop2_vectop_vsubc,
        rsp_cop2_vectop_vaddb,
        rsp_cop2_vectop_vsubb,
        rsp_cop2_vectop_vaccb,
        rsp_cop2_vectop_vsucb,
        rsp_cop2_vectop_vsad,
        rsp_cop2_vectop_vsac,
        rsp_cop2_vectop_vsum,
        rsp_cop2_vectop_vsaw,
        rsp_cop2_vectop_reserved,
        rsp_cop2_vectop_reserved,
        rsp_cop2_vectop_vlt,
        rsp_cop2_vectop_veq,
        rsp_cop2_vectop_vne,
        rsp_cop2_vectop_vge,
        rsp_cop2_vectop_vcl,
        rsp_cop2_vectop_vch,
        rsp_cop2_vectop_vcr,
        rsp_cop2_vectop_vmrg,
        rsp_cop2_vectop_vand,
        rsp_cop2_vectop_vnand,
        rsp_cop2_vectop_vor,
        rsp_cop2_vectop_vnor,
        rsp_cop2_vectop_vxor,
        rsp_cop2_vectop_vnxor,
        rsp_cop2_vectop_reserved,
        rsp_cop2_vectop_reserved,
        rsp_cop2_vectop_vrcp,
        rsp_cop2_vectop_vrcpl,
        rsp_cop2_vectop_vrcph,
        rsp_cop2_vectop_vmov,
        rsp_cop2_vectop_vrsq,
        rsp_cop2_vectop_vrsql,
        rsp_cop2_vectop_vrsqh,
        rsp_cop2_vectop_vnoop,
        rsp_cop2_vectop_vextt,
        rsp_cop2_vectop_vextq,
        rsp_cop2_vectop_vextn,
        rsp_cop2_vectop_reserved,
        rsp_cop2_vectop_vinst,
        rsp_cop2_vectop_vinsq,
        rsp_cop2_vectop_vinsn,
        rsp_cop2_vectop_reserved
};







void rsp_step()
{
        /* execute instruction */

        (instruction[__OPCODE])();



        /* calculate next pc */

        switch(rsp_reg.delay)
        {
            case EXEC_RSP_DELAY:   /* next instruction is at pc_delay */
                rsp_reg.delay = NO_RSP_DELAY;
                mem.sp_reg_pc = rsp_reg.pc_delay;
                break;

            case DO_RSP_DELAY:   /* a delayed pc was stored into pc_delay      */
                       /* next instr is at pc+4 (case 0 is executed) */
                       /* next instr is at pc_delay                  */
                rsp_reg.delay = EXEC_RSP_DELAY;

            case NO_RSP_DELAY:    /* normal execution */
                mem.sp_reg_pc += 4;

        } /* switch(rsp_reg.delay) */



        /* do common things */

        rsp_reg.r[0] = 0;
        rsp_reg.count++;

} /* void rsp_step() */





void rsp_dis(uWORD code)
{
        (instruction[code>>26])();
}


void rsp_reset()
{
        int i, j;



        /* debugging */

        rsp_reg.halt = 1;
        rsp_reg.count = 0;



        /* gpr registers */

        for(i=0; i<32; i++)
                rsp_reg.r[i] = 0;



        /* flags (ccr) */

        for(i=0; i<4; i++)
                rsp_reg.flag[i] = 0;



        /* vector registers */

        for(j=0; j<32; j++)
                for(i=0; i<8; i++)
                        rsp_reg.v[j][i] = 0;



        /* hidden registers */

        rsp_reg.vrcph_source = 0;
        rsp_reg.vrcph_result = 0;

} /* void rsp_reset() */



/******************************************************************************\
*                                                                              *
*   RSP (Reality Signal Processor)                                             *
*                                                                              *
\******************************************************************************/





static void rsp_special()
{
        (special_instruction[__F])();
}


static void rsp_regimm()
{
        (regimm_instruction[__RT])();
}


static void rsp_j()
{

        rsp_reg.delay = DO_RSP_DELAY;
        rsp_reg.pc_delay = __T;
BOX("j")
}





static void rsp_jal()
{
        rsp_reg.delay = DO_RSP_DELAY;
        rsp_reg.pc_delay = __T;
        rsp_reg.r[31] = (mem.sp_reg_pc + 8) & 0x1fff;
BOX("jal")
}





static void rsp_beq()
{
        if(rsp_reg.r[__RS] == rsp_reg.r[__RT])
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = __O;
        }
BOX("beq")
}





static void rsp_bne()
{
        if(rsp_reg.r[__RS] != rsp_reg.r[__RT])
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = __O;
        }
BOX("bne")
}






static void rsp_blez()
{
        if(((signed)rsp_reg.r[__RS]) <= 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = __O;
        }
BOX("blez")
}





static void rsp_bgtz()
{
        if(((signed)rsp_reg.r[__RS]) > 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = __O;
        }
BOX("bgtz")
}





static void rsp_addi()
{
        rsp_reg.r[__RT] = rsp_reg.r[__RS] + (sWORD)__I;
BOX("addi")
}





static void rsp_addiu()
{
        rsp_reg.r[__RT] = rsp_reg.r[__RS] + (sWORD)__I;
BOX("addiu")
}





static void rsp_slti()
{
        if(rsp_reg.r[__RS] < __I)
                rsp_reg.r[__RT] = 1;
        else
                rsp_reg.r[__RT] = 0;
BOX("slti")
}





static void rsp_sltiu()
{
        if((uWORD)rsp_reg.r[__RS] < (uWORD)__I)
                rsp_reg.r[__RT] = 1;
        else
                rsp_reg.r[__RT] = 0;
BOX("sltiu")
}





static void rsp_andi()
{
        rsp_reg.r[__RT] = rsp_reg.r[__RS] & (uHWORD)__I;
BOX("andi")
}





static void rsp_ori()
{
        rsp_reg.r[__RT] = rsp_reg.r[__RS] | (uHWORD)__I;
BOX("ori")
}





static void rsp_xori()
{
        rsp_reg.r[__RT] = rsp_reg.r[__RS] ^ (uHWORD)__I;
BOX("xori")
}





static void rsp_lui()
{
        rsp_reg.r[__RT] = __I << 16;
BOX("lui")
}


static void rsp_cop0()
{
        (cop0_instruction[__RS])();  
}





static void rsp_cop2()
{
        (cop2_instruction[__RS])();
}





static void rsp_lb()
{
        rsp_reg.r[__RT] = (sWORD)(sBYTE)((uBYTE *)mem.sp_dmem)[((rsp_reg.r[__RS]+__I)&0xfff)^3];
BOX("lb")
}





static void rsp_lh()
{
        uWORD addr;

        addr = rsp_reg.r[__RS]+__I;
        addr &= 0xfff;

        rsp_reg.r[__RT] = (sWORD)(sHWORD)
                            (mem.sp_dmem[(addr+0)^3] << 8)
                          | (mem.sp_dmem[(addr+1)^3]);
BOX("lh")
}





static void rsp_lw()
{
        uWORD addr;

        addr = rsp_reg.r[__RS]+__I;
        addr &= 0xfff;

        rsp_reg.r[__RT] = (sWORD)(sWORD)
                            (mem.sp_dmem[(addr+0)^3] << 24)
                          | (mem.sp_dmem[(addr+1)^3] << 16)
                          | (mem.sp_dmem[(addr+2)^3] << 8)
                          |  mem.sp_dmem[(addr+3)^3];
BOX("lw")
}





static void rsp_lbu()
{
        rsp_reg.r[__RT] = (sWORD)(uBYTE)((uBYTE *)mem.sp_dmem)[((rsp_reg.r[__RS]+__I)&0xfff)^3];
BOX("lbu")
}





static void rsp_lhu()
{
        uWORD addr;

        addr = rsp_reg.r[__RS]+__I;
        addr &= 0xfff;

        rsp_reg.r[__RT] = (sWORD)(uHWORD)
                            (mem.sp_dmem[(addr+0)^3] << 8)
                          | (mem.sp_dmem[(addr+1)^3]);
BOX("lhu")
}





static void rsp_sb()
{
        mem.sp_dmem[((rsp_reg.r[__RS]+__I)&0xfff)^3] = (uBYTE)rsp_reg.r[__RT];
BOX("sb")
}





static void rsp_sh()
{
        uWORD data, addr;

        data = rsp_reg.r[__RT];
        addr = rsp_reg.r[__RS]+__I;
        addr &= 0xfff;

        mem.sp_dmem[(addr+0)^3] = (uBYTE)(data >> 8);
        mem.sp_dmem[(addr+1)^3] = (uBYTE)(data);
BOX("sh")
}





static void rsp_sw()
{
        uWORD data, addr;

        data = rsp_reg.r[__RT];
        addr = rsp_reg.r[__RS]+__I;
        addr &= 0xfff;

        mem.sp_dmem[(addr+0)^3] = (uBYTE)(data >> 24);
        mem.sp_dmem[(addr+1)^3] = (uBYTE)(data >> 16);
        mem.sp_dmem[(addr+2)^3] = (uBYTE)(data >> 8);
        mem.sp_dmem[(addr+3)^3] = (uBYTE)(data >> 0);
BOX("sw")
}





static void rsp_lwc2()
{
        sWORD   addr;
        int     offset = __SA >> 1;
        int     i;

        switch(__RD)
        {
            case 0x00:
                addr = __RSP_O(0) + rsp_reg.r[__RS];
                addr &= 0xfff;

                if(offset & 1)
                {
                        rsp_reg.v[__RT][offset>>1] &= 0xff00;
                        rsp_reg.v[__RT][offset>>1] |= mem.sp_dmem[addr^3];
                }
                else
                {
                        rsp_reg.v[__RT][offset>>1] &= 0x00ff;
                        rsp_reg.v[__RT][offset>>1] |= mem.sp_dmem[addr^3] << 8;
                }
                break;

            case 0x01:
                addr = __RSP_O(1) + rsp_reg.r[__RS];
                addr &= 0xfff;

                rsp_reg.v[__RT][offset>>1] = (mem.sp_dmem[addr^3] << 8) | mem.sp_dmem[(addr+1)^3];
                break;

            case 0x02:
                addr = __RSP_O(2) + rsp_reg.r[__RS];
                addr &= 0xfff;

                rsp_reg.v[__RT][offset>>1] = (mem.sp_dmem[(addr)^3] << 8) | mem.sp_dmem[(addr+1)^3];
                rsp_reg.v[__RT][(offset>>1)+1] = (mem.sp_dmem[(addr+2)^3] << 8) | mem.sp_dmem[(addr+3)^3];
                break;

            case 0x03:
                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

                rsp_reg.v[__RT][offset>>1] = (mem.sp_dmem[(addr)^3] << 8) | mem.sp_dmem[(addr+1)^3];
                rsp_reg.v[__RT][(offset>>1)+1] = (mem.sp_dmem[(addr+2)^3] << 8) | mem.sp_dmem[(addr+3)^3];
                rsp_reg.v[__RT][(offset>>1)+2] = (mem.sp_dmem[(addr+4)^3] << 8) | mem.sp_dmem[(addr+5)^3];
                rsp_reg.v[__RT][(offset>>1)+3] = (mem.sp_dmem[(addr+6)^3] << 8) | mem.sp_dmem[(addr+7)^3];
                break;

            case 0x04:
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i < (16 - (addr & 0xf)); i++)
                {
                        if((offset+i) & 1)
                        {
                                rsp_reg.v[__RT][((offset+i)>>1)] &= 0xff00;
                                rsp_reg.v[__RT][((offset+i)>>1)] |= mem.sp_dmem[(addr+i)^3];
                        }
                        else
                        {
                                rsp_reg.v[__RT][((offset+i)>>1)] &= 0x00ff;
                                rsp_reg.v[__RT][((offset+i)>>1)] |= mem.sp_dmem[(addr+i)^3] << 8;
                        }
                }
                break;

            case 0x05:
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i < (addr & 0xf); i++)
                {
                        if(i & 1)
                        {
                                rsp_reg.v[__RT][7-(i>>1)] &= 0x00ff;
                                rsp_reg.v[__RT][7-(i>>1)] |= mem.sp_dmem[(addr-i-1)^3] << 8;
                        }
                        else
                        {
                                rsp_reg.v[__RT][7-(i>>1)] &= 0xff00;
                                rsp_reg.v[__RT][7-(i>>1)] |= mem.sp_dmem[(addr-i-1)^3];
                        }
                }
                break;

            case 0x06:
                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                rsp_reg.v[__RT][offset>>1] = mem.sp_dmem[addr^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+1] = mem.sp_dmem[(addr+1)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+2] = mem.sp_dmem[(addr+2)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+3] = mem.sp_dmem[(addr+3)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+4] = mem.sp_dmem[(addr+4)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+5] = mem.sp_dmem[(addr+5)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+6] = mem.sp_dmem[(addr+6)^3] << 8;
                rsp_reg.v[__RT][(offset>>1)+7] = mem.sp_dmem[(addr+7)^3] << 8;
                break;

            case 0x07:
                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                for(i=0; i<8; i++)
                {
                        rsp_reg.v[__RT][((offset)>>1)+i] = mem.sp_dmem[(addr+i)^3] << 7;
                }
                break;

            case 0x08:

                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                for(i=0; i < (16 - (addr & 0xf)); i+=2)
                {
                        rsp_reg.v[__RT][(offset+i)>>1] = mem.sp_dmem[(addr+i)^3] << 7;
                }
                break;

            case 0x09:
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i < (16 - (addr & 0xf)); i+=4)
                {
                        rsp_reg.v[__RT][(offset>>1)+(i>>2)] = mem.sp_dmem[(addr+i)^3] << 7;
                }
                break;

            case 0x0a:
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i<16; i++)
                {
                        if(i >= offset)
                        {
                                if(i & 1)
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0xff00;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[(addr+i-offset)^3];
                                }
                                else
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0x00ff;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[(addr+i-offset)^3] << 8;
                                }
                        }
                        else
                        {
                                if(i & 1)
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0xff00;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[(addr+i+16-offset)^3];
                                }
                                else
                                {
                                        rsp_reg.v[__RT][i>>1] &= 0x00ff;
                                        rsp_reg.v[__RT][i>>1] |= mem.sp_dmem[(addr+i+16-offset)^3] << 8;
                                }
                        }
                }
                break;

            case 0x0b:
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i<8; i++)
                {
                        rsp_reg.v[__RT+i][(8-(offset>>1)+i)&7] = (mem.sp_dmem[(addr+(i<<1))^3] << 8) | mem.sp_dmem[(addr+(i<<1)+1)^3];
                }
                break;

            default:
				break;
        } /* switch(__RD) */
BOX("lwc2")
}





static void rsp_swc2()
{
        sWORD   addr;
        int     offset = __SA >> 1;
        int     i;

        switch(__RD)
        {
            case 0x00:
                addr = __RSP_O(0) + rsp_reg.r[__RS];
                addr &= 0xfff;

                if(offset & 1)
                        mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1];
                else
                        mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1] >> 8;
                break;

            case 0x01:
                addr = __RSP_O(1) + rsp_reg.r[__RS];
                addr &= 0xfff;

                mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[(addr+1)^3] = rsp_reg.v[__RT][offset>>1];
                
                break;

            case 0x02:
                addr = __RSP_O(2) + rsp_reg.r[__RS];
                addr &= 0xfff;

                mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[(addr+1)^3] = rsp_reg.v[__RT][offset>>1];
                mem.sp_dmem[(addr+2)^3] = rsp_reg.v[__RT][(offset>>1)+1] >> 8;
                mem.sp_dmem[(addr+3)^3] = rsp_reg.v[__RT][(offset>>1)+1];
                break;

            case 0x03:
                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

                mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[(addr+1)^3] = rsp_reg.v[__RT][offset>>1];
                mem.sp_dmem[(addr+2)^3] = rsp_reg.v[__RT][(offset>>1)+1] >> 8;
                mem.sp_dmem[(addr+3)^3] = rsp_reg.v[__RT][(offset>>1)+1];
                mem.sp_dmem[(addr+4)^3] = rsp_reg.v[__RT][(offset>>1)+2] >> 8;
                mem.sp_dmem[(addr+5)^3] = rsp_reg.v[__RT][(offset>>1)+2];
                mem.sp_dmem[(addr+6)^3] = rsp_reg.v[__RT][(offset>>1)+3] >> 8;
                mem.sp_dmem[(addr+7)^3] = rsp_reg.v[__RT][(offset>>1)+3];
                break;

            case 0x04:
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i < (16 - (addr & 0xf)); i++)
                {
                        if((offset+i) & 1)
                                mem.sp_dmem[(addr+i)^3] = rsp_reg.v[__RT][(offset+i)>>1];
                        else
                                mem.sp_dmem[(addr+i)^3] = rsp_reg.v[__RT][(offset+i)>>1] >> 8;
                }
                break;

            case 0x05:
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i < (addr & 0xf); i++)
                {
                        if((offset+i) & 1)
                                mem.sp_dmem[(addr-i-1)^3] = rsp_reg.v[__RT][7-(i>>1)] >> 8;
                        else
                                mem.sp_dmem[(addr-i-1)^3] = rsp_reg.v[__RT][7-(i>>1)];
                }
                break;

            case 0x06:
                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                mem.sp_dmem[addr^3] = rsp_reg.v[__RT][offset>>1] >> 8;
                mem.sp_dmem[(addr+1)^3] = rsp_reg.v[__RT][(offset>>1)+1] >> 8;
                mem.sp_dmem[(addr+2)^3] = rsp_reg.v[__RT][(offset>>1)+2] >> 8;
                mem.sp_dmem[(addr+3)^3] = rsp_reg.v[__RT][(offset>>1)+3] >> 8;
                mem.sp_dmem[(addr+4)^3] = rsp_reg.v[__RT][(offset>>1)+4] >> 8;
                mem.sp_dmem[(addr+5)^3] = rsp_reg.v[__RT][(offset>>1)+5] >> 8;
                mem.sp_dmem[(addr+6)^3] = rsp_reg.v[__RT][(offset>>1)+6] >> 8;
                mem.sp_dmem[(addr+7)^3] = rsp_reg.v[__RT][(offset>>1)+7] >> 8;
                break;

            case 0x07:
                addr = __RSP_O(3) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                for(i=0; i<8; i++)
                {
                        mem.sp_dmem[(addr+i)^3] = rsp_reg.v[__RT][((offset)>>1)+i] >> 7;
                }
                break;

            case 0x08:
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                for(i=0; i < (16 - (addr & 0xf)); i+=2)
                {
                        mem.sp_dmem[(addr+i)^3] = rsp_reg.v[__RT][(offset+i)>>1] >> 7;
                }
                break;

            case 0x09:
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

    /* BUG: ??? offset should be removed ??? */
                for(i=0; i < (16 - (addr & 0xf)); i+=4)
                {
                        mem.sp_dmem[(addr+i)^3] = rsp_reg.v[__RT][(offset>>1)+(i>>2)] >> 7;
                }
                break;

            case 0x0a:
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                for(i=0; i<16; i++)
                {
                        if(i >= offset)
                        {
                                if(i & 1)
                                        mem.sp_dmem[(addr+i-offset)^3] = rsp_reg.v[__RT][i>>1];
                                else
                                        mem.sp_dmem[(addr+i-offset)^3] = rsp_reg.v[__RT][i>>1] >> 8;
                        }
                        else
                        {
                                if(i & 1)
                                        mem.sp_dmem[(addr+i+16-offset)^3] = rsp_reg.v[__RT][i>>1];
                                else
                                        mem.sp_dmem[(addr+i+16-offset)^3] = rsp_reg.v[__RT][i>>1] >> 8;
                        }
                }
                break;

            case 0x0b:
                addr = __RSP_O(4) + rsp_reg.r[__RS];
                addr &= 0xfff;

                /* The following was VERY VERY hard to figure out!!! */
                for(i=0; i<8; i++)
                {
                        sWORD   o = offset ? addr + (16-offset) : addr;
                        mem.sp_dmem[((o & ~0x0f) | ((o + (i<<1)) & 0x0f))^3] = rsp_reg.v[__RT+i][(8-(offset>>1)+i)&7] >> 8;
                        mem.sp_dmem[(((o & ~0x0f) | ((o + (i<<1)) & 0x0f)) + 1)^3] = rsp_reg.v[__RT+i][(8-(offset>>1)+i)&7];
                }
                break;

            default:
				break;
        } /* switch(__RD) */

BOX("swc2")
}






static void rsp_reserved()
{
}





static void rsp_special_sll()
{
        rsp_reg.r[__RD] = (uWORD)rsp_reg.r[__RT] << __SA;
//BOX("sll")
}





static void rsp_special_srl()
{
        rsp_reg.r[__RD] = (uWORD)rsp_reg.r[__RT] >> __SA;
BOX("srl")
}





static void rsp_special_sra()
{
        rsp_reg.r[__RD] = (sWORD)rsp_reg.r[__RT] >> __SA;
BOX("sra")
}





static void rsp_special_sllv()
{
        rsp_reg.r[__RD] = rsp_reg.r[__RT] << (rsp_reg.r[__RS] & 0x1f);
BOX("sllv")
}





static void rsp_special_srlv()
{
        rsp_reg.r[__RD] = (uWORD)rsp_reg.r[__RT] >> (rsp_reg.r[__RS] & 0x1f);
BOX("srlv")
}





static void rsp_special_srav()
{
        rsp_reg.r[__RD] = (sWORD)rsp_reg.r[__RT] >> (rsp_reg.r[__RS] & 0x1f);
BOX("srav")
}





static void rsp_special_jr()
{
        rsp_reg.delay = DO_RSP_DELAY;
        rsp_reg.pc_delay = 0x04000000 | (rsp_reg.r[__RS] & 0x1fff);
BOX("jr")
}





static void rsp_special_jalr()
{
        rsp_reg.delay = DO_RSP_DELAY;
        rsp_reg.pc_delay = 0x04000000 | (rsp_reg.r[__RS] & 0x1fff);
        rsp_reg.r[__RD] = (mem.sp_reg_pc + 8) & 0x1fff;
BOX("jalr")
}





static void rsp_special_break()
{
        ((uWORD *)mem.sp_reg)[4] |= 0x00000203;

        if(((uWORD *)mem.sp_reg)[4] & 0x40)
                //reg.do_or_check_sthg |= RS4300I_DO_SP_INTERRUPT;
				rs4300i_MI_Intr_SP();

        rsp_reg.halt = 1;
}





static void rsp_special_add()
{
        rsp_reg.r[__RD] = rsp_reg.r[__RS] + rsp_reg.r[__RT];
BOX("add")
}





static void rsp_special_addu()
{
        rsp_reg.r[__RD] = rsp_reg.r[__RS] + rsp_reg.r[__RT];
BOX("addu")
}





static void rsp_special_sub()
{
        rsp_reg.r[__RD] = rsp_reg.r[__RS] - rsp_reg.r[__RT];
BOX("sub")
}





static void rsp_special_subu()
{
        rsp_reg.r[__RD] = rsp_reg.r[__RS] - rsp_reg.r[__RT];
BOX("subu")
}





static void rsp_special_and()
{
        rsp_reg.r[__RD] = rsp_reg.r[__RS] & rsp_reg.r[__RT];
BOX("and")
}






static void rsp_special_or()
{
        rsp_reg.r[__RD] = rsp_reg.r[__RS] | rsp_reg.r[__RT];
BOX("or")
}





static void rsp_special_xor()
{
        rsp_reg.r[__RD] = rsp_reg.r[__RS] ^ rsp_reg.r[__RT];
BOX("xor")
}





static void rsp_special_nor()
{
        rsp_reg.r[__RD] = ~(rsp_reg.r[__RS] | rsp_reg.r[__RT]);
BOX("nor")
}





static void rsp_special_slt()
{
        if(rsp_reg.r[__RS] < rsp_reg.r[__RT])
                rsp_reg.r[__RD] = 1;
        else
                rsp_reg.r[__RD] = 0;
BOX("slt")
}





static void rsp_special_sltu()
{
        if((uWORD)rsp_reg.r[__RS] < (uWORD)rsp_reg.r[__RT])
                rsp_reg.r[__RD] = 1;
        else
                rsp_reg.r[__RD] = 0;
BOX("sltu")
}





static void rsp_special_reserved()
{
}




static void rsp_regimm_bltz()
{
        if(((sWORD)rsp_reg.r[__RS]) < 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = mem.sp_reg_pc + 4 + (__I << 2);
        }

BOX("bltz")
}





static void rsp_regimm_bgez()
{
        if(((sWORD)rsp_reg.r[__RS]) >= 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = mem.sp_reg_pc + 4 + (__I << 2);
        }
BOX("bgez")
}





static void rsp_regimm_bltzal()
{
        rsp_reg.r[31] = mem.sp_reg_pc + 8;

        if(((sWORD)rsp_reg.r[__RS]) < 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = mem.sp_reg_pc + 4 + (__I << 2);
        }
BOX("bltzal")
}





static void rsp_regimm_bgezal()
{
        rsp_reg.r[31] = mem.sp_reg_pc + 8;

        if(((sWORD)rsp_reg.r[__RS]) >= 0)
        {
                rsp_reg.delay = DO_RSP_DELAY;
                rsp_reg.pc_delay = mem.sp_reg_pc + 4 + (__I << 2);
        }
BOX("bgezal")
}





static void rsp_regimm_reserved()
{
}





static void rsp_cop0_mf()
{
        if(__RD < 8)
                rsp_reg.r[__RT] = doReadMemWord(0x04040000 | (__RD << 2));
        else
                rsp_reg.r[__RT] = doReadMemWord(0x04100000 | ((__RD & ~0x08) << 2));
  
BOX("mf")
}





static void rsp_cop0_mt()
{
        if(__RD < 8)
                doWriteMemWord(rsp_reg.r[__RT], 0x04040000 | (__RD << 2));
        else
                doWriteMemWord(rsp_reg.r[__RT], 0x04100000 | ((__RD & ~0x08) << 2));
BOX("mt")
}





static void rsp_cop0_reserved()
{
}





static void rsp_cop2_mf()
{
        sHWORD  data;
        uBYTE    offs;

        offs = (__SA >> 1) & 0x0f;

        if(offs & 0x01)   /* odd */
        {
                data =  rsp_reg.v[__RD][(offs >> 1) + 1] & 0x00ff;
                data += rsp_reg.v[__RD][offs >> 1]       & 0xff00;
        }
        else              /* even */
        {
                data = rsp_reg.v[__RD][offs >> 1];
        }

        rsp_reg.r[__RT] = data;
BOX("mf")
}





static void rsp_cop2_cf()
{
        rsp_reg.r[__RT] = rsp_reg.flag[__RD & 0x03];
}





static void rsp_cop2_mt()
{
        uHWORD   data;
        uBYTE    offs;

        data = rsp_reg.r[__RT];
        offs = (__SA >> 1) & 0x0f;

        if(offs & 0x01)   /* odd */
        {
                rsp_reg.v[__RD][offs >> 1]       &= 0xff00;
                rsp_reg.v[__RD][offs >> 1]       |= data >> 8;
                rsp_reg.v[__RD][(offs >> 1) + 1] &= 0x00ff;
                rsp_reg.v[__RD][(offs >> 1) + 1] |= data << 8;
        }
        else              /* even */
        {
                rsp_reg.v[__RD][offs >> 1] = data;
        }
BOX("mt")
}





static void rsp_cop2_ct()
{
        rsp_reg.flag[__RD & 0x03] = rsp_reg.r[__RT];  
}





static void rsp_cop2_vectop()
{
        (cop2_vectop_instruction[__F])();

}





static void rsp_cop2_reserved()
{
}





/* rsp cop2 vectop instructions */

static void rsp_cop2_vectop_vmulf()
{

        STORE_REGISTER
        VECT_ACCU_OP(ah, +, (uHWORD), *, >> 16)

        ACCU_2_VREG(ah);

//        PRINT_NOT_IMPLEMENTED("VMULF")

BOX("vmulf")
} /* static void rsp_cop2_vectop_vmulf() */





static void rsp_cop2_vectop_vmulu()
{
        PRINT_NOT_IMPLEMENTED("VMULU")

BOX("vmulu")
} /* static void rsp_cop2_vectop_vmulu() */





static void rsp_cop2_vectop_vrndp()
{
        PRINT_NOT_IMPLEMENTED("VRNDP")
BOX("vrndp")
} /* static void rsp_cop2_vectop_vrndp() */





static void rsp_cop2_vectop_vmulq()
{
        PRINT_NOT_IMPLEMENTED("VMULQ")
BOX("vmulq")
} /* static void rsp_cop2_vectop_vmulq() */





static void rsp_cop2_vectop_vmudl()
{
        STORE_REGISTER
        VECT_ACCU_OP_EQ(ah, (uHWORD), *, >> 16)

        ACCU_2_VREG(ah);
BOX("vmudl")
} /* static void rsp_cop2_vectop_vmudl() */





static void rsp_cop2_vectop_vmudm()
{
        STORE_REGISTER
        VECT_ACCU_OP_EQ(ah,(sHWORD), *, >> 16)

        ACCU_2_VREG(ah);
BOX("vmudm")
} /* static void rsp_cop2_vectop_vmudm() */





static void rsp_cop2_vectop_vmudn()
{

        STORE_REGISTER
        VECT_ACCU_OP_EQ(al, (uHWORD), *, )

        ACCU_2_VREG(al);
BOX("vmudn")
} /* static void rsp_cop2_vectop_vmudn() */





static void rsp_cop2_vectop_vmudh()
{
        STORE_REGISTER
        VECT_ACCU_OP_EQ(ah, (sHWORD), *, )

        ACCU_2_VREG(ah);
BOX("vmudh")
} /* static void rsp_cop2_vectop_vmudh() */





static void rsp_cop2_vectop_vmacf()
{
        STORE_REGISTER
        VECT_ACCU_OP(ah, +, (uHWORD), *, >> 16)

        ACCU_2_VREG(ah);

//        PRINT_NOT_IMPLEMENTED("VMACF")
BOX("vmacf")
} /* static void rsp_cop2_vectop_vmacf() */





static void rsp_cop2_vectop_vmacu()
{
        PRINT_NOT_IMPLEMENTED("VMACU")
BOX("vmacu")
} /* static void rsp_cop2_vectop_vmacu() */





static void rsp_cop2_vectop_vrndn()
{
        PRINT_NOT_IMPLEMENTED("VRNDN")
BOX("vrndn")
} /* static void rsp_cop2_vectop_vrndn() */





static void rsp_cop2_vectop_vmacq()
{
        PRINT_NOT_IMPLEMENTED("VMACQ")
BOX("vmacq")
} /* static void rsp_cop2_vectop_vmacq() */





static void rsp_cop2_vectop_vmadl()
{
        STORE_REGISTER
        VECT_ACCU_OP(ah, +, (uHWORD), *, >> 16)

        ACCU_2_VREG(ah);
BOX("vmadl")
} /* static void rsp_cop2_vectop_vmadl() */





static void rsp_cop2_vectop_vmadm()
{
        STORE_REGISTER
        VECT_ACCU_OP(ah, +, (sHWORD), *, >> 16)

        ACCU_2_VREG(ah);
BOX("vmadm")
} /* static void rsp_cop2_vectop_vmadm() */





static void rsp_cop2_vectop_vmadn()
{
        STORE_REGISTER
        VECT_ACCU_OP(al, +, (uHWORD), *, )

        ACCU_2_VREG(al);
BOX("vmadn")
} /* static void rsp_cop2_vectop_vmadn() */





static void rsp_cop2_vectop_vmadh()
{
        STORE_REGISTER
        VECT_ACCU_OP(ah, +, (sHWORD), *, )

        ACCU_2_VREG(ah);
BOX("vmadh")
} /* static void rsp_cop2_vectop_vmadh() */





static void rsp_cop2_vectop_vadd()
{
        STORE_REGISTER
        VECT_OP(+)
        ADD_CARRY(__SA)
        CLIP_RESULT_ADD
BOX("vadd")
} /* static void rsp_cop2_vectop_vadd() */





static void rsp_cop2_vectop_vsub()
{
        STORE_REGISTER
        VECT_OP(-)
        SUB_CARRY(__SA)
        CLIP_RESULT_SUB
BOX("vsub")
} /* static void rsp_cop2_vectop_vsub() */





static void rsp_cop2_vectop_vsut()
{
        PRINT_NOT_IMPLEMENTED("VSUT")
BOX("vsut")
} /* static void rsp_cop2_vectop_vsut() */





static void rsp_cop2_vectop_vabs()
{
        STORE_REGISTER
        VECT_OP_ABS
BOX("vabs")
} /* static void rsp_cop2_vectop_vabs() */





static void rsp_cop2_vectop_vaddc()
{
        STORE_REGISTER
        VECT_OP(+)
        SET_CARRY_ADD
BOX("vaddc")
} /* static void rsp_cop2_vectop_vaddc() */





static void rsp_cop2_vectop_vsubc()
{
        STORE_REGISTER
        VECT_OP(-)
        SET_CARRY_SUB
        SET_ZERO_FLAGS(__SA)
BOX("vmsubc")
} /* static void rsp_cop2_vectop_vsubc() */





static void rsp_cop2_vectop_vaddb()
{
        PRINT_NOT_IMPLEMENTED("VADDB")
BOX("vaddb")
} /* static void rsp_cop2_vectop_vaddb() */





static void rsp_cop2_vectop_vsubb()
{
        PRINT_NOT_IMPLEMENTED("VSUBB")
BOX("vsubb")
} /* static void rsp_cop2_vectop_vsubb() */





static void rsp_cop2_vectop_vaccb()
{
        PRINT_NOT_IMPLEMENTED("VACCB")
BOX("vaccb")
} /* static void rsp_cop2_vectop_vaccb() */





static void rsp_cop2_vectop_vsucb()
{
        PRINT_NOT_IMPLEMENTED("VSUCB")
BOX("vsucb")
} /* static void rsp_cop2_vectop_vsucb() */





static void rsp_cop2_vectop_vsad()
{
        PRINT_NOT_IMPLEMENTED("VSAD")
BOX("vsad")
} /* static void rsp_cop2_vectop_vsad() */





static void rsp_cop2_vectop_vsac()
{
        PRINT_NOT_IMPLEMENTED("VSAC")
BOX("vsac")
} /* static void rsp_cop2_vectop_vsac() */





static void rsp_cop2_vectop_vsum()
{
        PRINT_NOT_IMPLEMENTED("VSUM")
BOX("vsum")
} /* static void rsp_cop2_vectop_vsum() */





static void rsp_cop2_vectop_vsaw()
{
        PRINT_NOT_IMPLEMENTED("VSAW")
BOX("vsaw")
} /* static void rsp_cop2_vectop_vsaw() */





static void rsp_cop2_vectop_vlt()
{
        STORE_REGISTER
        VECT_OP_COMP(<)
        VECT_OP_COPY(rsp_reg.flag[1])
        rsp_reg.flag[0] = 0;
BOX("vlt")
} /* static void rsp_cop2_vectop_vlt() */





static void rsp_cop2_vectop_veq()
{
        STORE_REGISTER
        VECT_OP_COMP(==)
        VECT_OP_COPY(0)
        rsp_reg.flag[0] = 0;
BOX("veq")
} /* static void rsp_cop2_vectop_veq() */





static void rsp_cop2_vectop_vne()
{
        STORE_REGISTER
        VECT_OP_COMP(!=)
        VECT_OP_COPY(0xff)
        rsp_reg.flag[0] = 0;
BOX("vne")
} /* static void rsp_cop2_vectop_vne() */





static void rsp_cop2_vectop_vge()
{
        STORE_REGISTER
        VECT_OP_COMP(>=)
        VECT_OP_COPY(rsp_reg.flag[1])
        rsp_reg.flag[0] = 0;
BOX("vge")
} /* static void rsp_cop2_vectop_vge() */





static void rsp_cop2_vectop_vcl()
{
        PRINT_NOT_IMPLEMENTED("VCL")
BOX("vcl")
} /* static void rsp_cop2_vectop_vcl() */





static void rsp_cop2_vectop_vch()
{
        PRINT_NOT_IMPLEMENTED("VCH")
BOX("vch")
} /* static void rsp_cop2_vectop_vch() */





static void rsp_cop2_vectop_vcr()
{
        PRINT_NOT_IMPLEMENTED("VCR")
BOX("vcr")
} /* static void rsp_cop2_vectop_vcr() */





static void rsp_cop2_vectop_vmrg()
{
        int     i;

        STORE_REGISTER

        for(i=0; i<8; i++)
        {
                if( (1<<i) & rsp_reg.flag[1] )
                        rsp_reg.v[__SA][i] = rsp_reg.v_src1[i];
                else
                        rsp_reg.v[__SA][i] = rsp_reg.v_src2[i];
        }
BOX("vmrg")
} /* static void rsp_cop2_vectop_vmrg() */





static void rsp_cop2_vectop_vand()
{
        STORE_REGISTER
        VECT_OP(&)
BOX("vand")
} /* static void rsp_cop2_vectop_vand() */





static void rsp_cop2_vectop_vnand()
{
        STORE_REGISTER
        VECT_N_OP(&)
BOX("vnand")
} /* static void rsp_cop2_vectop_vnand() */





static void rsp_cop2_vectop_vor()
{
        STORE_REGISTER
        VECT_OP(|)
BOX("vor")
} /* static void rsp_cop2_vectop_vor() */





static void rsp_cop2_vectop_vnor()
{
        STORE_REGISTER
        VECT_N_OP(|)
BOX("vnor")
} /* static void rsp_cop2_vectop_vnor() */





static void rsp_cop2_vectop_vxor()
{
        STORE_REGISTER
        VECT_OP(^)
BOX("vxor")
} /* static void rsp_cop2_vectop_vxor() */





static void rsp_cop2_vectop_vnxor()
{
        STORE_REGISTER
        VECT_N_OP(^)
BOX("vnxor")
} /* static void rsp_cop2_vectop_vnxor() */





static void rsp_cop2_vectop_vrcp()
{
        rsp_reg.v[__SA][__RD & 0x07] = rsp_reg.vrcph_source;
        rsp_reg.vrcph_result = rsp_reg.v[__RT][__RS & 0x07];
BOX("vrcp")
} /* static void rsp_cop2_vectop_vrcp() */





static void rsp_cop2_vectop_vrcpl()
{
        double result;
        int    int_part;

        result = 0.4999963 / ((rsp_reg.vrcph_source << 16) | rsp_reg.v[__RT][__RS & 0x07]);
        
        rsp_reg.vrcph_result = (sHWORD)result;
        result = frexp(result, &int_part);
        if(result == 0)
                rsp_reg.v[__SA][__RD & 0x07] = 0;
        else
        {
                rsp_reg.v[__SA][__RD & 0x07] = (uHWORD)( ((*((DWORDLONG *)(&result)) & 0x000fffffffffffffUL) | 0x0010000000000000UL ) >> (abs(int_part) + 37) );
        }
BOX("vrcpl")
} /* static void rsp_cop2_vectop_vrcpl() */





static void rsp_cop2_vectop_vrcph()
{
        rsp_reg.v[__SA][__RD & 0x07] = rsp_reg.vrcph_result;
        rsp_reg.vrcph_source = rsp_reg.v[__RT][__RS & 0x07];
BOX("vrcph")
} /* static void rsp_cop2_vectop_vrcph() */





static void rsp_cop2_vectop_vmov()
{
        rsp_reg.v[__SA][__RD & 0x07] = rsp_reg.v[__RT][__RS & 0x07];
BOX("vmov")
} /* static void rsp_cop2_vectop_vmov() */





static void rsp_cop2_vectop_vrsq()
{
        PRINT_NOT_IMPLEMENTED("VRSQ")
BOX("vrsq")
} /* static void rsp_cop2_vectop_vrsq() */





static void rsp_cop2_vectop_vrsql()
{
        PRINT_NOT_IMPLEMENTED("VRSQL")
BOX("vrsql")
} /* static void rsp_cop2_vectop_vrsql() */





static void rsp_cop2_vectop_vrsqh()
{
        PRINT_NOT_IMPLEMENTED("VRSQH")
BOX("vrsqh")
} /* static void rsp_cop2_vectop_vrsqh() */





static void rsp_cop2_vectop_vnoop()
{
} /* static void rsp_cop2_vectop_vnoop() */





static void rsp_cop2_vectop_vextt()
{
        PRINT_NOT_IMPLEMENTED("VEXTT")
BOX("vextt")
} /* static void rsp_cop2_vectop_vextt() */





static void rsp_cop2_vectop_vextq()
{
        PRINT_NOT_IMPLEMENTED("VEXTQ")
BOX("vextq")
} /* static void rsp_cop2_vectop_vextq() */





static void rsp_cop2_vectop_vextn()
{
        PRINT_NOT_IMPLEMENTED("VEXTN")
BOX("vextn")
} /* static void rsp_cop2_vectop_vextn() */





static void rsp_cop2_vectop_vinst()
{
        PRINT_NOT_IMPLEMENTED("VINST")
BOX("vinst")
} /* static void rsp_cop2_vectop_vinst() */





static void rsp_cop2_vectop_vinsq()
{
        PRINT_NOT_IMPLEMENTED("VINSQ")
BOX("vinsq")
} /* static void rsp_cop2_vectop_vinsq() */





static void rsp_cop2_vectop_vinsn()
{
        PRINT_NOT_IMPLEMENTED("VINSN")
BOX("vinsn")
} /* static void rsp_cop2_vectop_vinsn() */





static void rsp_cop2_vectop_reserved()
{
        PRINT_NOT_IMPLEMENTED("RESERVED")
} /* static void rsp_cop2_vectop_reserved() */

